class DockerConfigException(Exception):
    pass


class DockerManagerException(Exception):
    pass


class DockerWorkerException(Exception):
    pass


class DockerArgsException(Exception):
    pass


class DockerApiClientException(Exception):
    pass


class DockerResourceCreationFailedException(Exception):
    pass
